/* Copyright (c) 2024, Yves Lepage */
#define MODULE "num_nlg"
#define TRACE 0
/* extern int tracet ; */

/*
 * Compute ordering of four positive real numbers.
 * Four possible orderings:
 *      0: for a:a::a:a
 *      1: for a:b::c:d
 *      2: for a:c::d:b
 *      3: for a:d::b:c
 *
 * Compute the analogical power.
 *
 * For instance:
 *     5 : 7 :: 10 : 12  =>  p =  1.0   arithmetic analogy
 *     2 : 4 ::  8 : 16  =>  p =  0.0   geometric analogy
 *   8/5 : 2 ::  4 :  8  =>  p = -1.0   harmonic analogy
 *     2 : 2 ::  3 :  4  =>  p = -100   minimum analogy
 *
 * Zero terms are also possible.
 *
 * For instance:
 *     0 : 3 :: 10 : 12  =>  p =  1.183
 *     0 : 0 :: 10 : 12  =>  p =  0.0
 *     0 : 0 ::  0 : 12  =>  p =  0.0
 *     0 : 0 ::  0 :  0  =>  p =  200.0   all R+ union +infinity
 *
 * Method:
 *  dichotomic search
 *
 * apart from the particular cases,
 * start with the two theoretical bounds of the analogical power
 * compute the generalized means of the extremes and the means
 * compare them
 * and recursively progress with the middle of the powers,
 * stop when the difference is small.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*
#include "usrlib.h"
#include "usrmath.h"
*/
#include "num_nlg.h"

/*
 * REPLACES #include "usrlib.h"
 * macro for trace:
 * use:
 * trace((string))
 * in C code not followed by semi-colon
 */

#define ytrace(x) { printf x ; fflush(stderr) ; fflush(stdout) ; }
#define ntrace(x)

#if TRACE
/* #define trace(x) { ytrace(x) ; if ( iscgiuse() ) printf("<br>\n") ; fflush(stdout) ; } */
#define trace(x) { ytrace(x) ; }
#else
#define trace(x)
#endif

/*
 * Exit on error
 */

#define ERR_MODULE_LABEL       " in module:   "
#define ERR_FUNCTION_LABEL     " in function: "
#define ERR_DIAGNOSTIC_LABEL   " diagnostic:  "

/*
 * Error, do not exit
 */

void warning(char *module_name, char *function_name, char *msg)
{
   fprintf(stderr,
          "\n#"
          "\n#" ERR_MODULE_LABEL     "%s"
          "\n#" ERR_FUNCTION_LABEL   "%s"
          "\n#" ERR_DIAGNOSTIC_LABEL "%s"
          "\n#"
          "\n",
          module_name,function_name,msg) ;
   fflush(stderr) ;
}

/*
 * Exit on error
 */

void error(char *module_name, char *function_name, char *msg)
{
    warning(module_name,function_name,msg) ;
    abort() ;
}


/*
 * Constants
 */

#define EPSILON          0.0001
#define MINUS_INFINITY  -100.0
#define PLUS_INFINITY   +100.0
#define ALL_R           -200.0
#define ALL_R_PLUS      +200.0
#define NO_SOLUTION     -300.0

/*
 * Trace macros
 */

#define F_NLGFMT    "%f : %f :: %f : %f" /* F_ for double numbers */
#define F_EQNFMT    "%f : %f ::^{%f} %f : x" /* F_ for double numbers */

/*
 * Local functions
 */

static int    isclose(double, double) ;

static void   swap(double *, double *) ;
static int    frame(double, double, double, double) ;
static int    static_ordering(double, double, double, double) ;

static double generalized_mean(double, double,
                              double) ;
static double difference(double, double, double, double,
                        double) ;

static double inner_dichotomic_search(double, double, double, double,
                                      double, double,
                                      double, double) ;
static double dichotomic_search(double, double, double, double) ;

static double static_power(double, double, double, double) ;
static void ordering_index_and_power(double, double, double, double,
                                     int *, double *) ;

/*
 * Almost equality.
 *
 * We use the formula implemented in numpy.isclose()
 * cf: https://numpy.org/doc/1.25/reference/generated/numpy.isclose.html
 *
 * An alternative formula is the one implemented in Python math.isclose()
 * cf: https://www.w3schools.com/python/ref_math_isclose.asp
 * double rel_tol = 1e-09, abs_tol = 0.0 ;
 * return ( fabs(a-b) < fmax(rel_tol * fmax(fabs(a), fabs(b)), abs_tol) ) ;
*/

static int isclose(double a, double b)
{
    double rtol = 1e-05, atol = 1e-08 ;

    return ( fabs(a-b) < atol + rtol * fabs(a+b) ) ;
}


/*
 * Enforce the terms to be positive non-null.
 * No more used. To be deleted in next version.
 */

void enforce_positive_terms(double *a, double *b, double *c, double *d)
{
    trace(("in  enforce_positive_terms(%f, %f, %f, %f)\n", *a, *b, *c, *d))

    if ( *a <= 0.0 ) *a = EPSILON ;
    if ( *b <= 0.0 ) *b = EPSILON ;
    if ( *c <= 0.0 ) *c = EPSILON ;
    if ( *d <= 0.0 ) *d = EPSILON ;

    trace(("out enforce_positive_terms(%f, %f, %f, %f)\n", *a, *b, *c, *d))
}

/*
 * Swap two numbers.
 */

static void swap(double *a, double *b)
{
    double tmp = *a ;

    *a = *b ;
    *b = tmp ;
}

/*
 * Sort four numbers.
 */

void sort_four_numbers(double *a, double *b, double *c, double *d)
{
    /* Hard-coded bubble sort */

trace(("in  sort_four_numbers(%f, %f, %f, %f)\n", *a, *b, *c, *d))

    if ( *a > *b ) swap(a, b) ;
    if ( *c > *d ) swap(c, d) ;
    if ( *a > *c ) swap(a, c) ;
    if ( *b > *d ) swap(b, d) ;
    if ( *b > *c ) swap(b, c) ;

trace(("out sort_four_numbers(%f, %f, %f, %f)\n", *a, *b, *c, *d))
}

/*
 * Check whether the extremes frame the means, or the opposite.
 */

static int frame(double a, double b, double c, double d)
{
    int result = 0 ;
    double *aa = &a, *dd = &d ;
    double *bb = &b, *cc = &c ;

trace(("in  frame(%f, %f, %f, %f) " F_NLGFMT "\n",
           a, b, c, d, *aa, *bb, *cc, *dd))

    if ( *dd < *aa )
        swap(aa, dd) ; /* Exchange of the extremes. */

trace(("mid 1 frame(%f, %f, %f, %f) " F_NLGFMT "\n",
           a, b, c, d, *aa, *bb, *cc, *dd))

    if ( *cc < *bb )
        swap(bb, cc) ; /* Exchange of the means. */

trace(("mid 2 frame(%f, %f, %f, %f) " F_NLGFMT "\n",
           a, b, c, d, *aa, *bb, *cc, *dd))

    if ( (*aa <= *bb) == (*cc <= *dd) )
        result = 1 ;
    else
        result = 0 ;

trace(("out frame(%f, %f, %f, %f) = %d\n", a, b, c, d, result))

        return result ;
}

/*
 * Compute the ordering of four real numbers.
 */

static int static_ordering(double a, double b, double c, double d)
{
    int result = -1 ;

trace(("in  static_ordering(%f, %f, %f, %f)\n", a, b, c, d))

    if ( a == b && b == c && c == d )
        result = 0 ; /* "a:a::a:a" */
    else
    {
        if ( frame(a, b, c, d) )
            result = 1 ; /* "a:b::c:d" */
        else
        {
trace(("mid 1 static_ordering(%f, %f, %f, %f)\n", a, b, c, d))
            if ( frame(a, c, d, b) )
                result = 2 ; /* "a:c::d:b" */
            else
            {
trace(("mid 2 static_ordering(%f, %f, %f, %f)\n", a, b, c, d))
                /* if ( frame(a, d, b, c) ) */
                    result = 3 ; /* "a:d::b:c" */
            } ;
        } ;
    } ;

trace(("out static_ordering(%f, %f, %f, %f) = %d\n", a, b, c, d, result))

    return result ;
}

static char *ordering_name(int ordering)
{
    char *result = (char *) calloc(9, sizeof(char)) ;

trace(("in  ordering_name(%d)\n", ordering))

    strncpy(result, "a:a::a:a\0", 9) ;
    if ( ordering == 1 )
        strncpy(result, "a:b::c:d\0", 9) ;
    else if ( ordering == 2 )
        strncpy(result, "a:c::d:b\0", 9) ;
    else if ( ordering == 3 )
        strncpy(result, "a:d::b:c\0", 9) ;

trace(("out ordering_name(%d) = %s\n", ordering, result))

    return result ;
}

char *_ordering_in_C(double a, double b, double c, double d)
{
    return (char *) ordering_name(static_ordering(a, b, c, d)) ;
}

/*
 * Compute the generalized mean of two non-negative real numbers.
 */

static double generalized_mean(double a, double d, double p)
{
    double result = 0.0 ;

trace(("in  generalized_mean(%f, %f, p=%f)\n", a, d, p))

    if (a == d)
        result = a ;
    else if ( p <= MINUS_INFINITY )
        result = fmin(a, d) ;
    else if ( 0.0 == p )
        result = pow((a * d), 0.5) ;
    else if ( p >= PLUS_INFINITY )
        result = fmax(a, d) ;
    else
    {
        if ( 0.0 == a )
            result = d * pow(0.5, 1/p) ;
        else if ( 0.0 == d )
            result = a * pow(0.5, 1/p) ;
        else
            result = d * pow(0.5 * (pow(a/d, p) + 1), 1/p) ;
    } ;

trace(("out generalized_mean(%f, %f, p=%f) = %f\n", a, d, p, result))

    return result ;
}

/*
 * Compute the difference of
 * the mean of the extremes and
 * the mean of the means
 * for an analogy between non-negative real numbers.
 */

static double difference(double a, double b, double c, double d, double p)
{
    double result = 0.0 ;

trace(("in  difference(" F_NLGFMT ", p=%f)\n", a, b, c, d, p))

    /* Error when the terms are not sorted in increasing order. */
    if ( ! ( a <= b && b <= c && c <= d ) )
        error(MODULE, "difference", "extremes and means do not frame one another.") ;
    result = generalized_mean(a, d, p) - generalized_mean(b, c, p) ;

trace(("out difference(" F_NLGFMT ", p=%f) = %f\n", a, b, c, d, p, result))

    return result ;
}

/*
 * Compute the analogical power between four ordered non-negative real numbers.
 * Recursive function.
 */

static double inner_dichotomic_search(double a, double b, double c, double d,
                                     double i, double j,
                                     double di, double dj)
{
    double result = 0 ;
    double m = 0.0 ;
    double dm = 0.0 ;

trace(("in  inner_dichotomic_search(" F_NLGFMT ",\n\t%f, %f,\n\t%f, %f)\n", a, b, c, d, i, j, di, dj))

    if ( ! ( i <= j ) )
    {
trace(("err inner_dichotomic_search(" F_NLGFMT ",\n\t%f, %f,\n\t%f, %f)\n", a, b, c, d, i, j, di, dj))
        error(MODULE, "inner_dichotomic_search", "i >= j.") ;
    } ;
    if ( isclose(i, j) ) {
        if ( fabs(di) < fabs(dj) )
            result = i ;
        else
            result = j ;
    }
    else
    {
        if ( ! ( di <= 0 && 0 <= dj ) )
        {
trace(("err inner_dichotomic_search(" F_NLGFMT ",\n\t%f, %f,\n\t%f, %f)\n", a, b, c, d, i, j, di, dj))
            error(MODULE, "inner_dichotomic_search", "di >= 0 or dj <= 0.") ;
        } ;
        m = (i+j) / 2 ;
        dm = difference(a, b, c, d, m) ;
trace(("mid inner_dichotomic_search(...)\n\t%f=%f < %f=%f < %f=%f\n",i,di,m,dm,j,dj))
        if ( 0.0 < dm )
            result = inner_dichotomic_search(a, b, c, d, i, m, di, dm) ;
        else
            result = inner_dichotomic_search(a, b, c, d, m, j, dm, dj) ;
    } ;

trace(("out inner_dichotomic_search(" F_NLGFMT ",\n\t%f, %f,\n\t%f, %f) = %f\n", a, b, c, d, i, j, di, dj, result))

    return result ;
}

/*
 * Compute the analogical power between four ordered non-negative real numbers.
 * This is only for the general case where p is not 0, or the infinities.
 *
 * CAUTION: IT IS ASSUMED THAT THE TERMS ARE SORTED IN INCREASING ORDER.
 */

static double dichotomic_search(double a, double b, double c, double d)
{
    double result = 0.0 ;
    double i  = MINUS_INFINITY, j  = PLUS_INFINITY ;
    double ii = MINUS_INFINITY, jj = PLUS_INFINITY ;
    double di = 0.0,            dj = 0.0 ;

trace(("in  dichotomic_search(" F_NLGFMT ") log(2)/log(a/b) = %f\n", a, b, c, d, log(2)/log(a/b)))

    /* Bounds for p, lower bound for i (ii) and upper bound for j (jj) */
    if ( 0.0 == a )
        ii = log(2) / log(d/b) - EPSILON ;
    else
        ii = log(2) / log(a/b) - EPSILON ;
    jj = log(2) / log(d/c) + EPSILON ;
    if ( i < ii )
        i = ii ;
    if ( i > PLUS_INFINITY )
        i = PLUS_INFINITY ;
    if ( jj < j )
        j = jj ;
    if ( j < MINUS_INFINITY )
        j = MINUS_INFINITY ;

trace(("mid dichotomic_search(" F_NLGFMT ")\n\t%f < p < %f\n", a, b, c, d, i, j))

    di = difference(a, b, c, d, i) ;
    dj = difference(a, b, c, d, j) ;

trace(("mid dichotomic_search(" F_NLGFMT ")\n\t%f < p < %f  di, dj = %f, %f\n", a, b, c, d, i, j, di, dj))

    result = inner_dichotomic_search(a, b, c, d, i, j, di, dj) ;

trace(("out dichotomic_search(" F_NLGFMT ") = %f\n", a, b, c, d, result))

    return result ;
}

/*
 * Compute the analogical power between four ordered non-negative real numbers.
 * Particular cases are dealt with first.
 * If not a particular case, call the dichotomic search.
 */

static double static_power(double a, double b, double c, double d)
{
    double result = 0.0 ;

trace(("in  static_power(" F_NLGFMT ")\n", a, b, c, d))

    if ( a == b && c == d )
    {
        if ( 0.0 == a && c == 0.0 )
        /* all values in R+ union plus infinity are possible */
            result = ALL_R_PLUS ;
        else
            /* all values in R union the infinities are possible */
            result = ALL_R ;
    }
    else if ( a == b ) /* but c != d */
    {
        if ( 0.0 == a )
            result = 0.0 ;
        else
            /*minimum analogy */
            result = MINUS_INFINITY ;
    }
    else if ( c == d ) /* but a != b */
        /* maximum analogy */
        result = PLUS_INFINITY ;
    else if ( isclose(a * d, b * c) )
        /* geometric analogy */
        result = 0.0 ;
    else if ( a - b == c - d )
        /* arithmetic analogy */
        result = 1.0 ;
    else if ( isclose((a + d) * (b * c), (b + c) * (a * d)) )
        /* harmonic analogy */
        result = -1.0 ;
    else
        result = dichotomic_search(a, b, c, d) ;

trace(("out static_power(" F_NLGFMT ") = %f\n", a, b, c, d, result))

    return result ;
}

double _power_in_C(double a, double b, double c, double d)
{
/*    enforce_positive_terms(&a, &b, &c, &d) ;*/
    sort_four_numbers(&a, &b, &c, &d) ;
    return static_power(a, b, c, d) ;
}

/*
 * Compute the solution of an analogical equation.
 */

double _solve_in_C(double a, double b, double c, double p)
{
    double result = NO_SOLUTION ;

trace(("in  _solve_in_C(" F_EQNFMT ")\n", a, b, p, c))

    if ( b > c )
        swap(&b, &c) ;
    if ( (p == ALL_R) && (a != b) && (a != c) )
        result = NO_SOLUTION ; /* No solution */
    else if ( isclose(a, b) )
        result = c ;
    else if ( isclose(a, c) )
        result = b ;
    else if ( MINUS_INFINITY >= p )
        result = c ;
    else if ( PLUS_INFINITY <= p )
        result = b ;
    else if ( isclose(0.0, p) )
    {
        if ( isclose(0.0, a) )
            result = PLUS_INFINITY ;
        else
            result = (b * c) / a ;
    }
    else
    {
        double ap = pow(a, p), means_p = pow(b, p) + pow(c, p) ;

        if ( ap > means_p )
            result = NO_SOLUTION ; /* No solution */
        else
            result = pow((means_p - ap), (1/p)) ;
    } ;

trace(("out _solve_in_C(" F_EQNFMT ") = %f\n", a, b, p, c, d))

    return result ;
}
