#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Interface de la bibliothèque pour l'analogie par moyenne généralisée.
Cas des nombres non-nuls positifs.

Seule cette interface est à utiliser dans les programmes Python:

>>> puissance(2, 1, 6, 3)
('a:b::c:d', 0.0)
>>> puissance(5, 7, 8, 6)
('a:d::b:c', 1.0)
>>> puissance(0, 3, 10, 12)
('a:b::c:d', 1.1829344593113598)

On peut calculer l'ordonnancement ou la puissance de façon isolée.
>>> puissance(2, 1, 6, 3, select='ord')
'a:b::c:d'
>>> puissance(5, 7, 8, 6, select='ord')
'a:d::b:c'
>>> puissance(2, 1, 6, 3, select='pui')
0.0
>>> puissance(5, 7, 8, 6, select='pui')
1.0
>>> puissance(5, 7, 8)
Traceback (most recent call last):
    ...
ValueError: Wrong number of argument(s) or wrong type of argument.

Le résultat pour une liste de listes est une liste de listes.
>>> puissance([[2, 1, 6, 3], [5, 7, 8, 6]])
(['a:b::c:d', 'a:d::b:c'], [0.0, 1.0])

Mais le résultat pour un n-uplet de n-uplets est un n-uplet de listes.
>>> puissance(((2, 1, 6, 3), (5, 7, 8, 6)))
(('a:b::c:d', 'a:d::b:c'), (0.0, 1.0))

Le résultat est un tableau Numpy pour un tableau Numpy.
>>> import numpy as np
>>> puissance(np.array([[2, 1, 6, 3], [5, 7, 8, 6]]))
(array(['a:b::c:d', 'a:d::b:c'], dtype='<U8'), array([0., 1.]))

Pour décomposer le résultat sur un tableau Numpy
en ses parties ordonnancements et puissances,
utiliser le tranchage [0 ou 1].  
>>> ps = puissance(np.array([[2, 1, 6, 3], [5, 7, 8, 6]]))
>>> print(ps[0])
['a:b::c:d' 'a:d::b:c']
>>> ps[0]
array(['a:b::c:d', 'a:d::b:c'], dtype='<U8')
>>> print(ps[1])
[0. 1.]
>>> ps[1]
array([0., 1.])

Le calcul isolé de l'ordonnancement ou de la puissance
sur une liste, un n-uplet ou un tableau Numpy est bien sûr possible.
>>> puissance([[2, 1, 6, 3], [5, 7, 8, 6]], select='pui')
[0.0, 1.0]

>>> puissance(np.array([[2, 1, 6, 3], [5, 7, 8, 6]]), select='ord')
array(['a:b::c:d', 'a:d::b:c'], dtype='<U8')

>>> résout(3, 4, 5, 1.0)
6
"""

#...!....1....!....2....!....3....!....4....!....5....!....6....!....7....!....8
################################################################################

__author__ = 'Yves Lepage <yves.lepage@waseda.jp>'
__date__, __version__ = '03/12/2024', '1.0' # Création
__date__, __version__ = '07/12/2024', '1.1' # Cas d'un terme nul.

################################################################################

import sys
import numpy as np
from _num_nlg import _ordering_in_C, _power_in_C, _solve_in_C

################################################################################

# La variable suivante est visible en dehors le paquetage.
ordres_possibles = ['a:a::a:a', 'a:b::c:d', 'a:c::d:b', 'a:d::b:c']

__valeur_ordonnancement__ = 'ordonnancement'
__valeur_puissance__ = 'puissance'

__valeur_ordering__ = 'ordering'
__valeur_power__ = 'power'

__valeurs__ = [__valeur_ordonnancement__, __valeur_puissance__,
                __valeur_ordering__, __valeur_power__]
__select_ordre__ = [__valeur_ordonnancement__, __valeur_ordering__]
__select_puissance__ = [__valeur_puissance__, __valeur_power__]

################################################################################

__erreur__dimension__ = "The last dimension of the input tensor must be 4 " \
                        "so as to be able to process analogies."
__erreur__argument__ = "Wrong number of argument(s) or wrong type of argument."
__erreur__equation__ = "Equation should have one unknown and only one."

################################################################################

def _is_prefix(s1, s2):
    """
    Tests whether the first string is a prefix of the second string.
    
    >>> _is_prefix('abc', 'abcdefg')
    True

    >>> _is_prefix(None, 'abcdefg')
    False
    >>> _is_prefix('', 'abcdefg')
    False
    >>> _is_prefix('abcdefg', 'abc')
    False
    >>> _is_prefix('abc', '')
    False
    """
    return s1 != None and len(s1) > 0 and s1 == s2[:len(s1)]

def _prefix_string(s1, s2s):
    """
    >>> _prefix_string('abc', ['abd', 'abde', 'abcde', 'abcdef'])
    'abcde'
    """
    for s2 in s2s:
        if _is_prefix(s1, s2):
            return s2

################################################################################

# Commentaire : de façon étonnante, plus rapide que _identité = lambda x: x...
def _identité(x):
    return x

def _my_raise(exception):
    # Astuce parce que `raise` est une déclaration,
    # mais une expression est requise après `else`.
    # Cf : https://stackoverflow.com/questions/54824572/raise-error-in-ternary-statement-in-python-without-using-classic-if-else-syntax
    raise exception

def _vectorisation_analogique(fonction):
    # On ne peut utiliser la fonction np.vectorize,
    #   car on joue sur des quadruplets.
    # On applique la fonction analogique le long de la dernière dimension,
    #   dont on s'assure qu'elle est bien 4,
    #   c'est-à-dire que l'on y a des quadruplets.
    # Il semble nécessaire de convertir les nombres en réels. Pourquoi ?
    return lambda tenseur: \
        np.apply_along_axis(lambda x: fonction(*x), -1, tenseur.astype(float)) \
            if tenseur.shape[-1] == 4 \
            else _my_raise(ValueError(__erreur__dimension__))

################################################################################

_ballage = {
    'quadruple': (_identité, _identité, _identité),
    'nparray':   (_identité, _vectorisation_analogique, _identité),
    'list':      (lambda x: np.array(x), _vectorisation_analogique, lambda x: x.tolist()),
    'tuple':     (lambda x: np.array(x), _vectorisation_analogique, lambda x: tuple(x.tolist())),
}

def _determine_type(*termes):
    if len(termes) == 1:
        if isinstance(*termes, np.ndarray):
            return 'nparray'
        elif isinstance(*termes, type([])):
            return 'list'
        elif isinstance(*termes, type(())):
            return 'tuple'
        else:
            return None
    elif len(termes) == 4:
        return 'quadruple'

def puissance(*termes, select=None):
    type_ = _determine_type(*termes)
    if type_ == None:
        raise ValueError(__erreur__argument__)
    emballage, vectorisation, déballage = _ballage[type_]

    if select == None:
        return déballage(vectorisation(_ordering_in_C)(*emballage(termes))), \
               déballage(vectorisation(_power_in_C)(*emballage(termes)))
    else:
        select = _prefix_string(select, __valeurs__)
        if select in __select_puissance__:
            return déballage(vectorisation(_power_in_C)(*emballage(termes)))
        elif select in __select_ordre__:
            return déballage(vectorisation(_ordering_in_C)(*emballage(termes)))

################################################################################

def résout(a=None, b=None, c=None, d=None, p=1.0):
    assert sum(X == None for X in (a, b, c, d)) == 1, \
        __erreur__equation__
    if   d == None:
        résultat = _solve_in_C(a, b, c, p)
    elif c == None:
        résultat = _solve_in_C(b, a, d, p)
    elif b == None:
        résultat = _solve_in_C(c, a, d, p)
    elif a == None:
        résultat = _solve_in_C(d, b, c, p)
    if résultat == -300:
        résultat = None
    return résultat

################################################################################

def _print_computations(tenseur):
    print('\nTenseur en entrée :\n')
    print(tenseur)
    print('\nTenseur des ordres :\n')
    print(puissance(tenseur, select='ord'))
    print('\nTenseur des puissances :\n')
    print(puissance(tenseur, select='puiss'))
    print('\nTenseur des ordres et puissances :\n')
    print(puissance(tenseur))

if __name__ == "__main__":
    # Comportement auprès de zéro.
    b, c, d = 5, 7, 9
    b, c, d = 3, 4, 5
    for a in [1, 0.1, 0.01, 0.001, 0.0001, 0]:
        print(f'{a} : {b} ::^{puissance(a, b, c, d)} {c} : {d}')

    # Un tenseur en Python convertit en tenseur Numpy.
    tenseur0 = [[[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]],
                [[11, 12, 13, 14], [15, 16, 17, 18], [19, 20, 21, 22]]]
    # Un tenseur de nombres entiers.
    tenseur1 = np.random.randint(1, 11, 48, dtype=int).reshape((2, 3, 2, 4))
    # Un tenseur de valeurs réelles dans [0,0; 1,0].
    tenseur2 = np.random.rand(60).reshape((3, 5, 4))

    # Application de la fonction puissance sur chque tenseur.
    for tenseur in [tenseur0, tenseur1, tenseur2]:
        _print_computations(tenseur)

